# Custom utils
import utils

import gc
import os
import json
import time
import random
import socket
import logging
import warnings
import datetime
import argparse
import numpy as np
from functools import partial

# Pytorch utils
import torch
from torch import cuda
from torch.backends import cudnn
import torch.distributed as dist
import torch.multiprocessing as mp
import torch.utils.data.distributed
from torch.utils.data import Dataset

# Dataset utils
from datasets import DataLoader

# Model utils
from transformers import AdamW, get_constant_schedule_with_warmup
from transformers import RobertaTokenizer, RobertaModel

# Bo utils
from botorch.models import SingleTaskGP
from botorch.optim import optimize_acqf
from botorch.fit import fit_gpytorch_model
from gpytorch.mlls import ExactMarginalLogLikelihood
from botorch.acquisition.logei import qLogExpectedImprovement

warnings.filterwarnings(action='ignore')
logging.getLogger("transformers.modeling_utils").setLevel(logging.ERROR)
silence_tensorflow()

large_task = ['sst2','qnli','qqp','mnli']
super_large_task = ['qnli','qqp', 'mnli']

class RobertaGLUE_base(torch.nn.Module):
    def __init__(self, config):
        super(RobertaGLUE_base, self).__init__()
        self.roberta = RobertaModel.from_pretrained(config.model_type, add_pooling_layer=False, ignore_mismatched_sizes=True)
        self.dense = torch.nn.Linear(config.hidden_size, config.hidden_size)
        self.dropout = torch.nn.Dropout(config.classifier_dropout)
        self.out = torch.nn.Linear(config.hidden_size, config.num_labels)
    
    def forward(self, input_ids, attention_mask):
        
        outputs = self.roberta(input_ids=input_ids, 
                               attention_mask=attention_mask)

        sequence_output = outputs[0]
        logits = sequence_output[:, 0, :]  
        logits = self.dropout(logits)
        logits = self.dense(logits)
        logits = torch.tanh(logits)
        logits = self.dropout(logits)  
        logits = self.out(logits)

        return logits
    
class RobertaGLUE(torch.nn.Module):
    def __init__(self, config):
        super(RobertaGLUE, self).__init__()
        self.roberta = RobertaModel.from_pretrained(config.model_type, add_pooling_layer=False, ignore_mismatched_sizes=True)
        
        for layer in self.roberta.encoder.layer[:config.freeze_num]:
            for param in layer.parameters():
                param.requires_grad = False
                
        self.second_part = torch.nn.ModuleList([*self.roberta.encoder.layer[config.freeze_num:]]) 
        self.dense = torch.nn.Linear(config.hidden_size, config.hidden_size)
        self.dropout = torch.nn.Dropout(config.classifier_dropout)
        self.out = torch.nn.Linear(config.hidden_size, config.num_labels)
        
    def forward(self, outputs):
        
        for layer in self.second_part:
            outputs = layer(outputs)
            outputs = outputs[0]
            
        sequence_output = outputs
        logits = sequence_output[:, 0, :]  
        logits = self.dropout(logits)
        logits = self.dense(logits)
        logits = torch.tanh(logits)
        logits = self.dropout(logits)
        logits = self.out(logits)

        return logits

class FeatureDataset(Dataset):
    def __init__(self, features, labels):
        self.features = features
        self.labels = labels   
        assert len(features) == len(labels), 'feature and label length should be same.'
    
    def __len__(self):
        return len(self.labels)
    
    def __getitem__(self, idx):
        return {'input_ids':self.features[idx], 
                'label':self.labels[idx]}
         
def generate_initial_points(bounds, n_points=3):
    lower_bounds, upper_bounds = bounds[0], bounds[1]
    scale = upper_bounds - lower_bounds
    points = lower_bounds + torch.rand((n_points, bounds.size(1))) * scale
    return points  

def validate(model,
             criterion,
             metric,
             valid_loader,
             device,
             freeze,
             task):
    
    model.eval()
    with torch.no_grad():
        
        valid_loss = 0.0
        valid_metric = 0.0
        for data in valid_loader:
            
            if freeze:
                input_ids = data['input_ids'].cuda(device)
                logits = model(input_ids)
                del input_ids
                
            else:
                input_ids = data['input_ids'].cuda(device)
                attention_mask = data['attention_mask'].cuda(device)
                logits = model(input_ids, attention_mask)
                del input_ids

            if task == 'stsb':
                targets = data['label'].to(torch.float32).cuda(device)
                logits = logits.to(torch.float32).cuda(device).squeeze()
            else:
                targets = data['label'].to(torch.int64).cuda(device)
                
            if criterion:
                loss = criterion(logits, targets)
                valid_loss += loss.item()
                
            _, add = metric.calculate(logits, targets)
            valid_metric += add  
            
            del add  
            del loss 
            del targets   
            del logits  
            gc.collect()
            torch.cuda.empty_cache()
    
    return valid_loss / len(valid_loader), valid_metric / len(valid_loader)

def train_iter(n_epochs,
               model,
               train_loader,
               valid_loader,
               task,
               criterion,
               metric,
               optimizer,
               lr_scheduler,
               freeze,
               gpu):
    
    stop_counter = 0
    best_metric = 0.0
    step = 0
    for epoch in range(n_epochs):
        
        if stop_counter > 4:
            break
        
        model.train()
        for data in train_loader:
            
            if step == 0:
                start_time = time.time()

            if freeze:
                input_ids = data['input_ids'].cuda(gpu)
                logits = model(input_ids)
                del input_ids
            
            else:
                input_ids = data['input_ids'].cuda(gpu)
                attention_mask = data['attention_mask'].cuda(gpu)
                logits = model(input_ids, attention_mask)
                del input_ids
                del attention_mask
            
            if task == 'stsb':
                targets = data['label'].to(torch.float64).cuda(gpu)
                logits = logits.to(torch.float64).cuda(gpu).squeeze()
            
            else:
                targets = data['label'].to(torch.int64).cuda(gpu)
            
            loss = criterion(logits, targets)
            
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()    
            lr_scheduler.step()
            
            if step == 0:
                elapse_time = time.time() - start_time
                elapse_time = datetime.timedelta(seconds=elapse_time)
                print(f'step time...{elapse_time}')
            
            del targets  
            del logits   
            del loss   
            
            gc.collect()
            torch.cuda.empty_cache()
            step += 1
        
        epoch_valid_loss, epoch_valid_metric = validate(model, criterion, metric, valid_loader, gpu, freeze, task)
        if gpu == 0:
            print(f"Epoch: {epoch+1} | valid loss: {epoch_valid_loss:.4f} | metric: {epoch_valid_metric:.4f}%")
        
        if epoch_valid_metric > best_metric:
            best_metric = epoch_valid_metric
            stop_counter = 0
        
        if epoch_valid_metric <= best_metric:
            stop_counter += 1  
    
    return np.float32(best_metric)

def find_free_port():
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(('', 0)) 
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]  

def cleanup():
    dist.destroy_process_group()
    
def main_worker(rank,
                world_size, 
                conn,
                config, 
                train_data,
                valid_data):
    
    utils.set_seed(config.train_seed)
    dist.init_process_group(backend='nccl', init_method=config.dist_url,
                            world_size=world_size, rank=rank)

    if (config.freeze_num != 0) and (config.task not in super_large_task):
        model = RobertaGLUE(config)
        freeze = True
    
    else:
        model = RobertaGLUE_base(config)
        freeze = False
        
    torch.cuda.set_device(rank)
    model.cuda(rank)

    config.batch_size = int(config.batch_size / world_size) #if ~(config.debug) else 1
    config.num_workers = int(config.num_workers / world_size) 
    
    model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[rank])
    criterion = torch.nn.CrossEntropyLoss() if config.task != 'stsb' else torch.nn.MSELoss()
    metric = utils.roberta_utils.metrics(config.task)

    train_sampler = torch.utils.data.distributed.DistributedSampler(train_data, num_replicas=world_size, rank=rank)
    train_loader = DataLoader(train_data,
                              batch_size=config.batch_size, 
                              num_workers=config.num_workers,
                              sampler=train_sampler,
                              worker_init_fn=utils.seed_worker,
                              shuffle=False, 
                              drop_last=False)
    
    valid_loader = DataLoader(valid_data,
                              batch_size=16, 
                              num_workers=config.num_workers,
                              shuffle=False, 
                              drop_last=False)

    total_steps = int(len(train_loader)*config.n_epochs)
    warmup_steps = int(total_steps*config.warmup_ratio)
    
    no_decay = ["bias", "LayerNorm.weight"]
    optimizer_grouped_parameters = [
        {
            "params": [p for n, p in model.named_parameters() if not any(nd in n for nd in no_decay)],
            "weight_decay": config.weight_decay
            },
        {
            "params": [p for n, p in model.named_parameters() if any(nd in n for nd in no_decay)],
            "weight_decay": 0.0
            }]
    
    optimizer = AdamW(optimizer_grouped_parameters,
                        lr=config.learning_rate,
                        eps=1e-06,
                        betas=(0.9,0.98))
    
    lr_scheduler = get_constant_schedule_with_warmup(
        optimizer,
        num_warmup_steps=warmup_steps
        )
    
    best_metric = train_iter(config.n_epochs,
                             model,
                             train_loader,
                             valid_loader,
                             config.task,
                             criterion,
                             metric,
                             optimizer,
                             lr_scheduler,
                             freeze,
                             rank)
    conn.send(best_metric)
        
    del model
    del optimizer
    del lr_scheduler
    
    cleanup()
    
def train(learning_rate, 
          batch_size, 
          config,
          train_data,
          valid_data):          
    
    # set hyperparameters
    config.batch_size = int(batch_size)
    config.learning_rate = float(learning_rate)
    
    # set other settings
    config.num_workers = 4
    config.world_size = torch.cuda.device_count()
    
    port = find_free_port()
    config.dist_url = f'tcp://localhost:{port}'
    
    parent_conn, child_conn = mp.Pipe()
    mp.spawn(main_worker, nprocs=config.world_size, args=(config.world_size, 
                                                          child_conn,
                                                          config,
                                                          train_data,
                                                          valid_data), join=True)
    best_metric = []
    while parent_conn.poll():
        best_metric.append(parent_conn.recv())
    
    gc.collect()
    torch.cuda.empty_cache()
    best_metric = best_metric[0]
    utils.set_seed(config.bo_seed)
    return best_metric

def optimize_hyperparameters(bounds, evaluate_model, config):
    
    train_x = generate_initial_points(bounds)
    train_y = torch.tensor([evaluate_model(learning_rate=lr, batch_size=bs) for lr, bs in train_x])

    for it in range(config.bo_iters):
        
        print(f">>>>>{it+1} iteration:")
                
        gp = SingleTaskGP(train_x, train_y.unsqueeze(-1))
        mll = ExactMarginalLogLikelihood(gp.likelihood, gp)
        fit_gpytorch_model(mll)
        
        acq_func = qLogExpectedImprovement(gp, best_f=train_y.max())

        candidates, _ = optimize_acqf(
            acq_function=acq_func,
            bounds=bounds,
            q=config.q,
            num_restarts=config.num_restarts,
            raw_samples=config.raw_samples,
        )

        new_x = candidates.detach()
        new_y = evaluate_model(learning_rate=new_x[..., 0], batch_size=new_x[..., 1])
        train_x = torch.cat([train_x, new_x])
        train_y = torch.cat([train_y, torch.tensor([new_y])])
        
        print(f'>>>>>> Selected learning rate: {train_x[-1][0]}')
        print(f'>>>>>> Selected batch size   : {int(train_x[-1][1])}')
        print(f'>>>>>> Validation Metric     : {train_y[-1]}')
        print(f'=====================================================')

    return train_x, train_y

def get_arg_parser():

    parser = argparse.ArgumentParser(description="outer bo glue")
    
    parser.add_argument('--task', type=str, default='rte')
    parser.add_argument('--n_epochs', type=int, default=10)
    parser.add_argument('--num_labels', type=int, default=2)
    parser.add_argument('--freeze_num', type=int, default=6)
    parser.add_argument('--weight_decay', type=float, default=0.1)
    parser.add_argument('--warmup_ratio', type=float, default=0.2)
    parser.add_argument('--learning_rate', type=float, default=None)
    parser.add_argument('--batch_size', type=int, default=None)
    
    parser.add_argument('--bo_seed', type=int, default=None)
    parser.add_argument('--train_seed', type=int, default=None)
    
    parser.add_argument('--model_type', type=str, default='roberta-base')
    parser.add_argument('--classifier_dropout', type=float, default=0.1)
    parser.add_argument('--hidden_size', type=int, default=768)
    
    parser.add_argument('--bo_iters', type=int, default=None)
    parser.add_argument('--q', type=int, default=1)
    parser.add_argument('--num_restarts', type=int, default=5)
    parser.add_argument('--raw_samples', type=int, default=20)
    
    parser.add_argument('--num_workers', type=int, default=4)
    parser.add_argument("--gpu_devices", type=int, nargs='+', default=None)
    parser.add_argument('--gpu', type=int, default=None)
    parser.add_argument('--rank', type=int, default=0)
    parser.add_argument('--world_size', type=int, default=1)
    parser.add_argument('--dist_url', type=str)
    
    return parser 

def main():
    
    parser = get_arg_parser()
    args = parser.parse_args()
    
    utils.set_seed(args.bo_seed) 
    utils.configure_cudnn(False)
    
    print(f'version name: {args.acq_version}')
    
    bounds = torch.tensor([[1e-6, 8], [1e-04, 16]], dtype=torch.float32) if args.task not in large_task else\
        torch.tensor([[1e-7, 32], [1e-04, 64]], dtype=torch.float32)
    
    print("Loading datasets...")
    if (args.freeze_num > 0) and (args.task not in super_large_task):
        train_data = torch.load(f'./data/{args.task}_train_features.pt')
        valid_data = torch.load(f'./data/{args.task}_validation_matched_features.pt') if args.task == 'mnli' else\
            torch.load(f'./data/{args.task}_validation_features.pt')
        
    else:
        train_data = torch.load(f'./data/{args.task}_train.pt')
        valid_data = torch.load(f'./data/{args.task}_validation_matched.pt') if args.task == 'mnli' else\
            torch.load(f'./data/{args.task}_validation.pt')
    
    evaluate_model = partial(train,
                             config=args,
                             train_data=train_data,
                             valid_data=valid_data)
    
    print(f"Starting outer BO with {args.task} task...")
    start_time = time.time()
    optimized_params = optimize_hyperparameters(bounds, evaluate_model, args)
    elapse_time = time.time() - start_time
    elapse_time = datetime.timedelta(seconds=elapse_time)
    
    print(f'Total time cost: {elapse_time}')
    hps, results = optimized_params
    max_value = torch.max(results)
    max_hps = hps[torch.argmax(results)].tolist()
    print(f'Max valid metric result: {max_value}')
    print(f'-> Hyperparameters: {max_hps}')
    
    log_data = {
        'cost':f'{elapse_time}',
        'learning_rate':hps[:,0].tolist(),
        'batch_size':hps[:,1].tolist(),
        'metric':results.tolist()
    }
    with open(f'./logs/freeze_{args.freeze_num}_{args.task}_outer_bo.json', 'w') as file:
        json.dump(log_data, file)
    
if __name__ == '__main__':
    main()