import gc
import os
import json
import time
import random
import socket
import logging
import warnings
import datetime
import argparse
import numpy as np
from functools import partial

# Pytorch utils
import torch
from torch import cuda
from torch.backends import cudnn
import torch.distributed as dist
import torch.multiprocessing as mp
import torch.utils.data.distributed
from torch.utils.data import Dataset

# Dataset utils
import datasets
from datasets import load_metric

# Model utils
from transformers import get_linear_schedule_with_warmup
from transformers import T5Tokenizer, T5ForConditionalGeneration
from transformers import AdamW

# Bo utils
from botorch.models import SingleTaskGP
from botorch.optim import optimize_acqf
from botorch.fit import fit_gpytorch_model
from gpytorch.mlls import ExactMarginalLogLikelihood
from botorch.acquisition.logei import qLogExpectedImprovement

import os 
import torch
from tqdm import tqdm
from collections import Counter

import re
import string

import numpy as np

def generate_summary(model, tokenizer, source, source_mask, target):
  
  output = model.module.generate(source,
                max_length=16, 
                early_stopping=True)
  
  target[target[:,:]==-100] = 0
  
  machine_text = [tokenizer.decode(senten, skip_special_tokens=True) for senten in output]
  human_text = [tokenizer.decode(senten, skip_special_tokens=True) for senten in target]
  machine_text = ["replacement_word" if text == "" else text for text in machine_text]
  human_text = ["replacement_word" if text == "" else text for text in human_text]
  
  del source
  del output 
  del target

  return {"machine_text": machine_text, "human_text":human_text}

def normalize_answer(s):
    """Lower text and remove punctuation, articles and extra whitespace."""
    def remove_articles(text):
        return re.sub(r'\b(a|an|the)\b', ' ', text)

    def white_space_fix(text):
        return ' '.join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return ''.join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(s))))

def f1_score(prediction, ground_truth):
    prediction_tokens = normalize_answer(prediction).split()
    ground_truth_tokens = normalize_answer(ground_truth).split()
    
    common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
    num_same = sum(common.values())
    if num_same == 0:
        return 0
    precision = 1.0 * num_same / len(prediction_tokens)
    recall = 1.0 * num_same / len(ground_truth_tokens)
    f1 = (2 * precision * recall) / (precision + recall)
    return f1

def exact_match_score(prediction, ground_truth):
    return (normalize_answer(prediction) == normalize_answer(ground_truth))

def metric_max_over_ground_truths(metric_fn, prediction, ground_truths):
    return metric_fn(prediction, ground_truths)

def qa_metrics(gold_answers, predictions):
    f1 = exact_match = total = 0

    for ground_truths, prediction in zip(gold_answers, predictions):
      total += 1
      exact_match += metric_max_over_ground_truths(
                    exact_match_score, prediction, ground_truths)
      f1 += metric_max_over_ground_truths(
          f1_score, prediction, ground_truths)
    
    exact_match = 100.0 * exact_match / total
    f1 = 100.0 * f1 / total
    return {'f1':f1, 'em':exact_match}

def evaluate(model,
             tokenizer,
             valid_loader,
             metric,
             device):
    
    em = 0.0
    f1 = 0.0
    with torch.no_grad():
        for i, batch in enumerate(tqdm(valid_loader)):
            
                source, source_mask, target, _ = torch.tensor(batch['input_ids']).to(device),\
                    batch['attention_mask'].to(device),\
                    batch['labels'], batch['decoder_attention_mask'].to(device)
                target = torch.tensor(target)
                target[target[: ,:] == 0 ] = -100
                target = target.to(device)
                
                all_text = generate_summary(model, tokenizer, source, source_mask, target)
                result = qa_metrics(all_text['machine_text'], all_text['human_text'])
                
                em += result['em']
                f1 += result['f1']

    print(f'Exact matching score: {em/len(valid_loader)}')
    print(f'F1 score: {f1/len(valid_loader)}')
    return (em + f1) / len(valid_loader)
 
def train_fn(model,
          tokenizer,
          optimizer,
          scheduler,
          train_loader, 
          valid_loader,
          metric,
          accumulation_steps=None,
          device=None):
        
    epoch_train_loss = 0.0
    for i, batch in enumerate(tqdm(train_loader)):
        
        source, source_mask, target, _ = torch.tensor(batch['input_ids']).to(device),\
            batch['attention_mask'].to(device),\
            batch['labels'],\
            batch['decoder_attention_mask'].to(device)
            
        target = torch.tensor(target)
        target[target[: ,:] == 0 ] = -100 # to let the model skip computing the loss of the zeros (paddings)
        target = target.to(device)
        
        output = model(source, source_mask, labels = target, return_dict=True)
        loss = output['loss']
        loss = loss / accumulation_steps 
        loss.backward()
        
        if (i+1) % accumulation_steps == 0:
            optimizer.step()
            optimizer.zero_grad()
            scheduler.step()
        
        epoch_train_loss += loss.item()
           
    epoch_train_loss = epoch_train_loss / len(train_loader)
    epoch_rouge_f1 = evaluate(model, tokenizer, valid_loader, metric, device)
    
    return epoch_train_loss, epoch_rouge_f1

warnings.filterwarnings(action='ignore')
logging.getLogger("transformers.modeling_utils").setLevel(logging.ERROR)
    
def set_seed(seed):
    os.environ['PYTHONHASHSEED'] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

def seed_worker(seed):
    worker_seed = torch.initial_seed() % 2**32
    random.seed(worker_seed)
    np.random.seed(worker_seed)
    torch.manual_seed(worker_seed)
    torch.cuda.manual_seed(worker_seed)
    torch.cuda.manual_seed_all(worker_seed)

def configure_cudnn(debug):
    cudnn.enabled = True
    cudnn.benchmark = True
    if debug:
        cudnn.deterministic = True
        cudnn.benchmark = False

def cleanup():
    dist.destroy_process_group()

def get_arg_parser():
    parser =  argparse.ArgumentParser(description=" A parser for T5 HP-BO")
    
    # Model configs
    parser.add_argument("--model", type=str, default='t5-base')
    parser.add_argument('--freeze_num', type=int, default=0)
    
    # Training configs
    parser.add_argument("--lr", type=float, default=None)
    parser.add_argument("--epoch", type=int, default=None)
    parser.add_argument("--metric", type=str, default=None)
    parser.add_argument('--train_seed', type=int, default=None)
    parser.add_argument("--batch_size", type=int, default=None)
    parser.add_argument('--ckpt_path', type=str, default="./ckpt")
    
    # DDP configs
    parser.add_argument('--rank', type=int)
    parser.add_argument('--world_size', type=int)
    parser.add_argument('--gpu', type=int)
    parser.add_argument('--num_workers', type=int, default=4)
    parser.add_argument("--gpu_devices", type=int, nargs='+')
    parser.add_argument('--distributed', action='store_true')
    parser.add_argument('--dist-backend', default='nccl', type=str)
    parser.add_argument('--dist-url', default='tcp://localhost:32465', type=str)
    
    # BO configs
    parser.add_argument('--q', type=int)
    parser.add_argument('--bo_seed', type=int)
    parser.add_argument('--bo_iters', type=int)
    parser.add_argument('--raw_samples', type=int)
    parser.add_argument('--num_restarts', type=int)
    
    return parser

def main_worker(rank, world_size, conn, args):

    set_seed(args.train_seed)
    configure_cudnn(False)

    os.makedirs(args.ckpt_path, exist_ok=True)
    
    print("Use GPU: {} for training...".format(rank))

    dist.init_process_group(backend='nccl', init_method=args.dist_url,
                            world_size=world_size, rank=rank)
    
    args.batch_size = int(args.batch_size / world_size)
    args.num_worker = int(args.num_workers / world_size)

    metric = load_metric(args.metric)
    
    print("Loading {} Model...".format(args.model))
    model = T5ForConditionalGeneration.from_pretrained(args.model)
    tokenizer = T5Tokenizer.from_pretrained(args.model, use_fast=True)
    
    if args.freeze_num > 0:
        top_layers = len(model.encoder.block)
        for layers in model.encoder.block[top_layers-6:]:
            for param in layers.parameters():
                param.requires_grad = False    
                
    cuda.set_device(rank)
    model.cuda(rank)
    
    no_decay = ["bias", "LayerNorm.weight"]
    optimizer_grouped_parameters = [
        {
            "params": [p for n, p in model.named_parameters() if not any(nd in n for nd in no_decay)],
            "weight_decay": 0.1,
            },
        {
            "params": [p for n, p in model.named_parameters() if any(nd in n for nd in no_decay)],
            "weight_decay": 0.0,
            },
        ]
    
    optimizer = AdamW(optimizer_grouped_parameters, 
                      lr=args.lr)
    
    print('Loading Data Loaders...')
    class SquadV2Dataset(Dataset):
        def __init__(self, tokenizer, max_length, split):
            self.dataset = datasets.load_dataset('squad_v2', split=split)
            self.tokenizer = tokenizer
            self.max_length = max_length

        def __len__(self):
            return len(self.dataset)

        def __getitem__(self, idx):
            item = self.dataset[idx]
            context = item['context']
            question = item['question']
            answers = item['answers']['text'][0] if item['answers']['text'] else ''

            input_text = f"question: {question} context: {context}"
            answer_text = answers

            input_encoding = self.tokenizer(
                input_text,
                max_length=self.max_length,
                padding='max_length',
                truncation=True,
                return_tensors="pt"
            )
            answer_encoding = self.tokenizer(
                answer_text,
                max_length=10,  
                padding='max_length',
                truncation=True,
                return_tensors="pt"
            )

            return {
                'input_ids': input_encoding['input_ids'].squeeze(),
                'attention_mask': input_encoding['attention_mask'].squeeze(),
                'labels': answer_encoding['input_ids'].squeeze(),
                'decoder_attention_mask': answer_encoding['attention_mask'].squeeze()
            }
    
    train_ds = SquadV2Dataset(tokenizer, 512, split='train')
    valid_ds = SquadV2Dataset(tokenizer, 512, split='validation')

    train_sampler = torch.utils.data.distributed.DistributedSampler(train_ds)
    train_loader = torch.utils.data.DataLoader(dataset=train_ds, 
                                               batch_size=args.batch_size, 
                                               sampler=train_sampler, 
                                               worker_init_fn=seed_worker)
    valid_loader = torch.utils.data.DataLoader(dataset= valid_ds, batch_size=32)

    scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=int(len(train_loader) * 0.1),
        num_training_steps=int(len(train_loader) * args.epoch))
    
    model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[rank])
    num_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    print("The number of parameters of model is {}...".format(num_params))
        
    epoch_start = time.time(); best_metric = 0.0
    for ep in range(args.epoch):
        epoch_loss, epoch_metric = train_fn(model,
                                            tokenizer,  
                                            optimizer,
                                            scheduler,
                                            train_loader, 
                                            valid_loader,
                                            metric,
                                            2,
                                            rank)
                      
        elapse_time = time.time() - epoch_start
        elapse_time = datetime.timedelta(seconds=elapse_time)
        
        if rank == 0:
            print(f"Epoch: {ep+1} | train loss: {epoch_loss:.4f} | valid metrics: {epoch_metric:.4f}% | time: {elapse_time}")

        if epoch_metric > best_metric:
            best_metric = epoch_metric
 
    conn.send(best_metric)
    
    del model
    del optimizer
    del scheduler
    
    cleanup()

def find_free_port():
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(('', 0)) 
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]  
    
def train(learning_rate,
          batch_size,
          args):
    
    args.learning_rate = learning_rate
    args.batch_size = int(batch_size)

    args.world_size = torch.cuda.device_count()
    
    port = find_free_port()
    args.dist_url = f'tcp://localhost:{port}'
    
    parent_conn, child_conn = mp.Pipe()
    mp.spawn(main_worker, nprocs=args.world_size, args=(args.world_size, child_conn, args), join=True)
    
    best_metric = []
    while parent_conn.poll():
        best_metric.append(parent_conn.recv())
    
    gc.collect()
    torch.cuda.empty_cache()
    best_metric = best_metric[0]
    set_seed(args.bo_seed)
    return torch.tensor(best_metric, dtype=torch.float32)

def generate_initial_points(bounds, n_points):
    lower_bounds, upper_bounds = bounds[0], bounds[1]
    scale = upper_bounds - lower_bounds
    points = lower_bounds + torch.rand((n_points, bounds.size(1))) * scale
    return points  

def optimize_hyperparameters(bounds, evaluate_model, config):
    
    train_x = generate_initial_points(bounds)
    train_y = torch.tensor([evaluate_model(learning_rate=lr, batch_size=bs) for lr, bs in train_x])

    for it in range(config.bo_iters):
        
        print(f">>>>>{it+1} iteration:")
                
        gp = SingleTaskGP(train_x, train_y.unsqueeze(-1))
        mll = ExactMarginalLogLikelihood(gp.likelihood, gp)
        fit_gpytorch_model(mll)
        
        acq_func = qLogExpectedImprovement(gp, best_f=train_y.max())
            
        candidates, _ = optimize_acqf(
            acq_function=acq_func,
            bounds=bounds,
            q=config.q,
            num_restarts=config.num_restarts,
            raw_samples=config.raw_samples,
        )

        new_x = candidates.detach()
        new_y = evaluate_model(learning_rate=new_x[..., 0], batch_size=new_x[..., 1])
        train_x = torch.cat([train_x, new_x])
        train_y = torch.cat([train_y, torch.tensor([new_y])])
        
        print(f'>>>>>> Selected learning rate: {train_x[-1][0]}')
        print(f'>>>>>> Selected batch size   : {int(train_x[-1][1])}')
        print(f'>>>>>> Validation Metric     : {train_y[-1]}')
        print(f'=====================================================')

    return train_x, train_y

def main():
    
    parser = get_arg_parser()
    args = parser.parse_args()
    
    set_seed(args.bo_seed)
    
    bounds = torch.tensor([[1e-6, 32], [1e-04, 64]], dtype=torch.float32) 
    evaluate_model = partial(train, args=args)
    
    print(f"Starting outer BO with {args.task} task...")
    start_time = time.time()
    optimized_params = optimize_hyperparameters(bounds, evaluate_model, args)
    elapse_time = time.time() - start_time
    elapse_time = datetime.timedelta(seconds=elapse_time)
    
    print(f'Total time cost: {elapse_time}')
    hps, results = optimized_params
    max_value = torch.max(results)
    max_hps = hps[torch.argmax(results)].tolist()
    print(f'Max valid metric result: {max_value}')
    print(f'-> Hyperparameters: {max_hps}')
    
    log_data = {
        'cost':f'{elapse_time}',
        'learning_rate':hps[:,0].tolist(),
        'batch_size':hps[:,1].tolist(),
        'metric':results.tolist()
    }
    with open(f'./logs/freeze_{args.freeze_num}_squad_outer_bo.json', 'w') as file:
        json.dump(log_data, file)
    
if __name__ =="__main__":
    main()